@@ -6,11 +6,17 @@ import base64
 import logging
 import itertools
 from UserDict import DictMixin
+LOG = logging.getLogger(__name__)
 
 def unpickle(f):
     if isinstance(f, basestring): f = open(f, 'rb')
     return pickle.load(f)
 
+def save_pickle(obj, filename):
+    with open_for_atomic_overwrite(filename) as f:
+        pickle.dump(obj, f, -1)
+        return f.tell()
+
 def get_picklecached_thing(filename, func=None, name=None):
     # This functionality is superceded by PickleDict.get_lazy.
     if name is None: name = filename
@@ -21,16 +27,15 @@ def get_picklecached_thing(filename, func=None, name=None):
         opener = open
     try:
         f = opener(filename, 'rb')
-        print 'Loading', name
+        LOG.info('Loading %s', name)
         result = unpickle(f)
         f.close()
     except IOError:
         if func is None: raise
-        print 'Computing', name
+        LOG.info('Computing %s', name)
         result = func()
-        print 'Saving', name
-        with open_for_atomic_overwrite(filename) as f:
-            pickle.dump(result, f, -1)
+        LOG.info('Saving %s', name)
+        save_pickle(result, filename)
     return result
 load_pickle = get_picklecached_thing
 
@@ -91,14 +96,6 @@ def human_readable_size(sz, multiplier=1000, sizes=['B', 'kB', 'MB', 'GB']):
         sz = sz / float(multiplier)
     return '%.2f%s' % (sz, name)
 
-def get_ipython_history(num_entries=15):
-    from IPython import ipapi
-    ip = ipapi.get()
-    if ip is None: return None
-    input_lines = [item.strip() for item in ip.user_ns['In']]
-    return [item for item in input_lines
-            if item and not item.startswith('?') and not item.endswith('?')][-num_entries:]
-    
 
 class PickleDict(object, DictMixin):
     '''
@@ -108,7 +105,7 @@ class PickleDict(object, DictMixin):
 
     >>> import tempfile
     >>> dirname = tempfile.mkdtemp()
-    >>> pd = PickleDict(dirname)
+    >>> pd = PickleDict(dirname, extension='.pkl')
 
     Let's clear out the directory so the tests start from a known state.
     
@@ -121,7 +118,7 @@ class PickleDict(object, DictMixin):
     It keeps an internal cache, so to make sure it's actually storing
     persistently, let's make a new one.
 
-    >>> pd = PickleDict(dirname)
+    >>> pd = PickleDict(dirname, extension='.pkl')
     >>> pd['abc']
     123
 
@@ -131,6 +128,12 @@ class PickleDict(object, DictMixin):
     ['abc']
     >>> pd.items()
     [('abc', 123)]
+    >>> 'abc' in pd
+    True
+    >>> '' in pd
+    False
+    >>> 'blue' in pd
+    False
 
     If you're just using string keys, you can use the item-to-attr
     adaptor `d`:
@@ -176,6 +179,12 @@ class PickleDict(object, DictMixin):
     >>> pd['dir2'][7, 8]
     (9, 10)
 
+    ...and delete subdirectories also:
+
+    >>> del pd['dir2']
+    >>> 'dir2' in pd
+    False
+
     It can also lazily compute an expensive function only if the
     result is not already pickled. (This replaces
     get_picklecached_thing.)
@@ -199,9 +208,6 @@ class PickleDict(object, DictMixin):
     >>> pd['_meta']['the_answer']['type'] == str(int)
     True
 
-    If you were running from within IPython, the metadata dict also
-    includes 'context', which holds your last 15 IPython inputs.
-
     There's a friendlier interface to getting metadata too:
 
     >>> pd.get_meta('the_answer', 'type') == str(int)
@@ -214,15 +220,16 @@ class PickleDict(object, DictMixin):
     
     '''
     special_character = '+'
-    __slots__ = ['logger', 'log', 'dir', 'gzip', 'store_metadata', 'history_len', 'cache']
+    __slots__ = ['logger', 'log', 'dir', 'store_metadata', 'cache', 'extension', 'load_pickle', 'save_pickle']
     
-    def __init__(self, dir, gzip=False, store_metadata=True, log=True, context_history_len=20):
+    def __init__(self, dir, store_metadata=True, log=True, extension='', load_pickle=load_pickle, save_pickle=save_pickle):
         self.logger = logging.getLogger('csc_utils.persist.PickleDict')
         self.log = log
         self.dir = os.path.abspath(os.path.expanduser(dir))
-        self.gzip = gzip
+        self.extension = extension
         self.store_metadata = store_metadata
-        self.history_len = context_history_len
+        self.load_pickle = load_pickle
+        self.save_pickle = save_pickle
         if not os.path.isdir(self.dir):
             os.makedirs(self.dir)
         self.clear_cache()
@@ -235,7 +242,7 @@ class PickleDict(object, DictMixin):
 
     @property
     def d(self): return ItemToAttrAdaptor(self)
-    
+
     def path_for_key(self, key):
         # FIXME: This method is hugely inefficient but for the moment necessary.
         #
@@ -253,6 +260,8 @@ class PickleDict(object, DictMixin):
         if not isinstance(key, basestring) or key.startswith(self.special_character) or '/' in key:
             for filename in os.listdir(self.dir):
                 if not filename.startswith(self.special_character): continue
+                if self.extension and filename.endswith(self.extension):
+                    filename = filename[:-len(self.extension)]
                 if self.key_for_path(filename) == key:
                     key = filename
                     break
@@ -262,6 +271,8 @@ class PickleDict(object, DictMixin):
         return os.path.join(self.dir, key)
 
     def key_for_path(self, path):
+        if self.extension and path.endswith(self.extension):
+            path = path[:-len(self.extension)]
         if path.startswith(self.special_character):
             return pickle.loads(base64.urlsafe_b64decode(path[1:]))
         return path
@@ -280,16 +291,16 @@ class PickleDict(object, DictMixin):
         if os.path.isdir(path):
             # Keep sub-PickleDict objects in cache, so that they
             # can cache their own data.
-            return PickleDict(path, gzip=self.gzip, store_metadata=self.store_metadata)
+            return PickleDict(path, store_metadata=self.store_metadata,
+                              extension=self.extension,
+                              load_pickle=self.load_pickle,
+                              save_pickle=self.save_pickle)
+        # Otherwise, expect an actual pickle object, so use the extension.
+        path = path + self.extension
         if not os.path.exists(path):
             raise KeyError(key)
         if self.log: self.logger.info('Loading %r...', key)
-        try:
-            import gzip
-            data = unpickle(gzip.open(path))
-        except IOError:
-            data = unpickle(open(path, 'rb'))
-
+        data = self.load_pickle(path)
         if self.log: self.logger.info('Loaded %r (%s).', key, type(data))
         return data
         
@@ -303,49 +314,44 @@ class PickleDict(object, DictMixin):
             return data
     
     def __setitem__(self, key, val):
-        if self.log: self.logger.info('Saving %r...', key)
+        if self.log: self.logger.info('Saving %r... (%s)', key, type(val))
         self.cache[key] = val
-        with open_for_atomic_overwrite(self.path_for_key(key)) as f:
-            if self.gzip:
-                import gzip
-                f = gzip.GzipFile(fileobj=f)
-            pickle.dump(val, f, -1)
-            if self.gzip:
-                f.close()
-            if self.log: self.logger.info('Saved %r (%s)', key, human_readable_size(f.tell()))
+        size = self.save_pickle(val, self.path_for_key(key) + self.extension)
+        if self.log:
+            if isinstance(size, int):
+                self.logger.info('Saved %r (%s)', key, human_readable_size(size))
+            else:
+                self.logger.info('Saved %r', key)
 
         if self.store_metadata:
             meta = {}
             meta['type'] = str(type(val))
-
-            # Add IPython history
-            try:
-                meta['context'] = get_ipython_history(num_entries=self.history_len)
-            except ImportError:
-                pass
             self['_meta'][key] = meta
 
     def __delitem__(self, key):
+        path = self.path_for_key(key)
+        if os.path.isdir(path):
+            self[key]._clear()
+            remaining = os.listdir(path)
+            if remaining:
+                raise RuntimeError(
+                    "PickleDict was trying to delete the subdirectory %s, but "
+                    "it still has the following files in it: %r"
+                    % (path, remaining))
+            else:
+                os.rmdir(path)
+        else:
+            os.remove(path + self.extension)
         self.cache.pop(key, None) # don't fail if it's not cached.
-        os.remove(self.path_for_key(key))
-
-
-    def _nuke(self):
-        '''
-        Destroy the directory tree, aka, ``rm -rf``.
-        '''
-        import shutil
-        shutil.rmtree(self.dir)
-        
         
     def _clear(self):
         '''
         Clear everything in the directory.
         '''
-        self._nuke()
-        os.makedirs(self.dir)
-        self.clear_cache()
-
+        for key in self:
+            del self[key]
+        if '_meta' in self:
+            del self['_meta']
     
     def clear(self):
         raise NotImplementedError("`clear`? Do you really mean that? If so, run _clear instead.")
@@ -371,7 +377,13 @@ class PickleDict(object, DictMixin):
         return self[name]
 
     def rename(self, old, new):
-        os.rename(self.path_for_key(old), self.path_for_key(new))
+        old_path = self.path_for_key(old)
+        new_path = self.path_for_key(new)
+        if not os.path.isdir(old_path):
+            old_path = old_path + self.extension
+            new_path = new_path + self.extension
+        os.rename(old_path, new_path)
+
         if old in self.cache:
             if not isinstance(self.cache[old], PickleDict):
                 self.cache[new] = self.cache[old]
@@ -387,7 +399,12 @@ class PickleDict(object, DictMixin):
         return list(self.__iter__())
 
     def has_key(self, key):
-        return (self.store_metadata and key == '_meta') or os.path.exists(self.path_for_key(key))
+        if key == '':
+            return False # Otherwise would match the directory itself.
+        if self.store_metadata and key == '_meta':
+            return True
+        path = self.path_for_key(key)
+        return os.path.exists(path) or os.path.exists(path+self.extension)
 
     def get_meta(self, key, meta_key, default_value=None):
         try:
@@ -530,7 +547,7 @@ class PickleDict(object, DictMixin):
 
 class MetaPickleDict(PickleDict):
     def __init__(self, dir):
-        super(MetaPickleDict, self).__init__(dir, gzip=False, store_metadata=False, log=False)
+        super(MetaPickleDict, self).__init__(dir, store_metadata=False, log=False)
         
 
 ###
